package cd.casic.ci.log.service;

import cd.casic.ci.log.dal.dao.IndexMapper;
import cd.casic.framework.redis.core.RedisTemplateUtils;
import cd.casic.ci.log.dal.dataobject.LogIndicesV2;
import com.github.benmanes.caffeine.cache.Caffeine;
import com.github.benmanes.caffeine.cache.LoadingCache;
import jakarta.annotation.Resource;
import lombok.extern.slf4j.Slf4j;
import org.apache.logging.log4j.util.Strings;
import org.redisson.RedissonRedLock;
import org.redisson.api.RLock;
import org.redisson.api.RedissonClient;
import org.springframework.stereotype.Service;

import java.util.concurrent.TimeUnit;

/**
 * @author by mianbin
 * @Classname IndexService
 * @Description TODO
 * @Date 2025/3/20 10:53
 */
@Slf4j
@Service
public class IndexService {

    @Resource
    private IndexMapper indexMapper;

    @Resource
    private RedissonClient redissonClient;

    @Resource
    private RedisTemplateUtils redisTemplateUtils;

    private static String LOG_INDEX_LOCK = "log:build:enable:lock:key";
    private static String LOG_LINE_NUM = "log:build:line:num:";
    private static String LOG_LINE_NUM_LOCK = "log:build:line:num:distribute:lock";
    private static Long INDEX_CACHE_MAX_SIZE = 100000L;
    private static Long INDEX_CACHE_EXPIRE_MINUTES = 30L;
    private static Long INDEX_LOCK_EXPIRE_SECONDS = 10L;

    // 返回 redis Key
    public static String getLineNumRedisKey(String buildId) {
        return LOG_LINE_NUM + buildId;
    }

    // Caffeine 本地数据库，不存在数据时自动添加
    private LoadingCache<String, String> indexCache = Caffeine.newBuilder()
            .maximumSize(INDEX_CACHE_MAX_SIZE)
            .expireAfterAccess(INDEX_CACHE_EXPIRE_MINUTES, TimeUnit.MINUTES)
            .build(key -> initialIndex(key));

    // 如果不存在初始化 index 方法
    private String initialIndex(String buildId) {
        String indexName = indexMapper.getIndexName(buildId);

        // 如果数据库里面都不存在 indexName，就使用分布式锁在数据库里面新增
        if (Strings.isBlank(indexName)){
            String lockKey = LOG_INDEX_LOCK + ":" + buildId;
            RLock lock = redissonClient.getLock(lockKey);
            RedissonRedLock redLock = new RedissonRedLock(lock);
            try{
                redLock.lock();
                // 二次验证，防止重复
                 indexName = indexMapper.getIndexName(buildId);
                if (Strings.isBlank(indexName)){
                    log.info(buildId + " Add the build record");
                    indexName = saveIndex(buildId);
                }
            } finally {
                lock.unlock();
            }
        }
        return indexName;
    }

    // 保存index到 数据库 和 Redis 里面去，最后返回值保存在 Caffeine 里面
    private String saveIndex(String buildId) {
        String indexName = indexMapper.getIndexName(buildId);
        // 保存到 mongo
        indexMapper.create(buildId,indexName,true);
        // 保存到 redis
        redisTemplateUtils.set(getLineNumRedisKey(buildId),"1",TimeUnit.DAYS.toSeconds(2));

        log.info("["+buildId+"|"+indexName+"] Create new index in db and cache");

        return indexName;
    }

    public String getIndexName(String buildId) {
        // 这里的 get 如果不存在就会初始化调用 initialIndex 方法
        String index = indexCache.get(buildId);
        if (Strings.isBlank(index)) {
            log.error("Fail to get the index of build from caffeine cache:" + buildId);
            index = "testId";
        }
        return index;
    }

    public long getAndAddLineNum(String buildId, int size) {
        // 获得并发锁
        RLock lock = redissonClient.getLock("$LOG_LINE_NUM_LOCK:$buildId");
        RedissonRedLock redissonRedLock = new RedissonRedLock(lock);
        try {
            redissonRedLock.lock();
            Object lineNum = redisTemplateUtils.get(getLineNumRedisKey(buildId));
            // 缓存命中则直接进行自增，缓存未命中则从db中取值，自增后再刷新缓存
            if (lineNum == null) {
                log.warn("[" + buildId + "|[" + size + "] Fail to get and add the line num, get from db");
                // 从 indexMapper 里面取出
                LogIndicesV2 build = indexMapper.getBuild(buildId);
                if (build == null) {
                    log.warn("[" + buildId + "|[" + size + "] The build is not exist in db");
                    return -1;
                }
                log.warn("[" + buildId + "|" + size + "] Got from db, lastLineNum: " + build.getLastLineNum());
                lineNum = build.getLastLineNum() + size;
                redisTemplateUtils.set(getLineNumRedisKey(buildId), lineNum, TimeUnit.DAYS.toSeconds(2));
                return build.getLastLineNum();
            } else {
                lineNum = redisTemplateUtils.incr(getLineNumRedisKey(buildId), size);
            }
            return (long) lineNum - size;
        } finally {
            redissonRedLock.unlock();
        }
    }

    public String getBuildIndexName(String buildId){
        LogIndicesV2 logIndices = indexMapper.getBuild(buildId);
        if (logIndices == null) {
            return "";
        }
        return logIndices.getIndexName();
    }

    public long getLastLineNum(String buildId) {
        Object number = redisTemplateUtils.get(getLineNumRedisKey(buildId));
        if (number != null) {
            return (long) number;
        }
        log.error("Redis can't find the lineNumber");

        // 如果为空就从数据库里面去取
        LogIndicesV2 build = indexMapper.getBuild(buildId);
        if (build != null) {
            return build.getLastLineNum();
        }

        // 如果还为空就返回0
        return 0;
    }

    // 清理缓存并保存行数
    public void flushLineNum2DB(String buildId) {
        String lineNum = (String) redisTemplateUtils.get(getLineNumRedisKey(buildId));
        if (Strings.isBlank(lineNum)) {
            log.warn("[$buildId] Fail to get lineNum from redis");
            return;
        }
        // 看数据是否合法
        long latestLineNum = 0;
        try {
            latestLineNum = Long.parseLong(lineNum);
        } catch (Exception e) {
            log.warn("[" + buildId + "|" + lineNum + "] Fail to convert line num to long", e);
            return;
        }
        // 如果更新成功
        if (indexMapper.updateLastLineNum(buildId, latestLineNum)) {
            redisTemplateUtils.del(getLineNumRedisKey(buildId));
        } else {
            log.warn("[" + buildId + "|" + latestLineNum + "] Fail to update the build latest line num");
        }
    }


}
